@@ -1,6 +1,6 @@
 #| bootstrap for rep.lang.interpreter
 
-   $Id: interpreter.jl,v 1.11 2001/04/30 20:35:22 jsh Exp $
+   $Id: interpreter.jl,v 1.14 2001/08/08 06:00:20 jsh Exp $
 
    Copyright (C) 2000 John Harper <john@dcs.warwick.ac.uk>
 
@@ -39,7 +39,9 @@
 (make-binding-immutable '#F)
 (make-binding-immutable '#T)
 
-(export-bindings '(nil t #F #T))
+(%define #undefined '#undefined)
+
+(export-bindings '(nil t #F #T #undefined))
 
 
 ;; function syntax
@@ -169,7 +171,7 @@ functions."
 	     (cond ((consp x) (car x))
 		   (t x))) bindings)
    (mapcar (lambda (x)
-	     (cond ((consp x) (cons 'setq x))
+	     (cond ((consp x) (list 'setq (car x) (cons 'progn (cdr x))))
 		   (t (list 'setq x nil)))) bindings)))
 
 (defmacro let-fluids (bindings . body)
@@ -298,14 +300,7 @@ string associated with VARIABLE."
      (defvar ,var nil ,doc)
      (setq ,var ,value)))
 
-;; XXX backwards compatibility crap
-(defmacro define-value (var-form value)
-  (if (eq (car var-form) 'quote)
-      ;; constant symbol
-      (list '%define (nth 1 var-form) value)
-    (error "define-value must take a constant variable")))
-
-(export-bindings '(setq-default define-special-variable define-value))
+(export-bindings '(setq-default define-special-variable))
 
 
 ;; Misc syntax
@@ -493,7 +488,11 @@ DATA)' while the handler is evaluated (these are the arguments given to
   (write t (format nil "*** %s: %s"
 		   (or (get err 'error-message) err)
 		   (mapconcat (lambda (x)
-				(format nil "%s" x)) data ", "))))
+				(format nil "%s" x)) data ", ")))
+  ;; XXX ugh.. so kludgey..
+  (open-structures '(rep.lang.error-helper))
+  (declare (bound error-helper))
+  (error-helper err data))
 
 (defvar error-handler-function default-error-handler)
 
@@ -556,6 +555,7 @@ of THUNK) each function will be called exactly once."
 (defmacro eval-when-compile (form)
   "FORM is evaluated at compile-time *only*. The evaluated value is inserted
 into the compiled program. When interpreted, nil is returned."
+  (declare (unused form))
   nil)
 
 ;; Hide interactive decls
